'----------------------------------------------------------------------------------------------------
'
' MCP23S17_COG v1.00 - MCP23S17 Parallel I/O expander driver
'
' This device driver supports up to eight MCP23S17 devices on one SPI bus for up to 128 bits of I/O!
'
' The low level SPI code runs at 4MBPS in order to use as little memory as possible.
'
' You will need the Microchip data sheet for the MCP23S17
'
' Note: IOCON=0 in the debugger, however if you set IOCON=1, the register map is
'       compatible with the MCP23S08
'
' Copyright July 23, 2010 by William Henning
'
' This code is released under the terms of the MIT license, attached below
'
' You will be able to find the latest version of this code at my site:
'
' http://Mikronauts.com/downloads
'

'----------------------------------------------------------------------------------------------------
{{
┌──────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────┐
│                                                   TERMS OF USE: MIT License                                                  │
├──────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────┤
│Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation    │
│files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,    │
│modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software│
│is furnished to do so, subject to the following conditions:                                                                   │
│                                                                                                                              │
│The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.│
│                                                                                                                              │
│THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE          │
│WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR         │
│COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,   │
│ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                         │
└──────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────┘
}}

'----------------------------------------------------------------------------------------------------
' Constants defined for MCP23S17_COG
'----------------------------------------------------------------------------------------------------

CON

  '--------------------------------------------------------------------------------------------------
  ' MCP23S17_COG command constants
  '--------------------------------------------------------------------------------------------------

  PIOREAD       = 1
  PIOWRITE      = 2

  '--------------------------------------------------------------------------------------------------
  ' SPI pin constants
  '--------------------------------------------------------------------------------------------------

  ' PropCade & Mem+ common usage

  CS            = 0             ' 0 if on P0
  CLK           = 1             ' 1 if on P1
  MOSI          = 2             ' 2 if on P2
  MISO          = 3             ' 3 if on P3

  ' PropCade also uses P9/P10/P11 as address inputs to a demultiplexer which is enabled by /CS above
  PIO_DEVICE    = 6             ' only used on PropCade

  '--------------------------------------------------------------------------------------------------
  ' PIO SPI commands
  '--------------------------------------------------------------------------------------------------

  PIOREADK      = %0100_000_1_00000000
  PIOWRITEK     = %0100_000_0_00000000

'----------------------------------------------------------------------------------------------------
' Start MCP23S17_COG
'
' You must reserve 4 longs for the mailbox, and pass its address to Start in order to run MCP23S17_COG.
'
'----------------------------------------------------------------------------------------------------

VAR

  long cmdptr, addrptr, dataptr


PUB start(mailbox)

  cmdptr   := mailbox
  addrptr  := mailbox+4
  dataptr  := mailbox+8

  word[cmdptr]   := $FFFF0000

  cognew(@MCP23S17_COG,mailbox)

  repeat while long[cmdptr]     ' should fix startup bug heater found - it was the delay to load/init the cog

PUB rdpbyte(n,vaddr)
  repeat while long[cmdptr]
  long[cmdptr] := (((PIOREADK + (n << 9))+vaddr)<<9)|PIOREAD
  repeat while long[cmdptr]
  return long[dataptr]

PUB wrpbyte(n,vaddr,val)
  repeat while long[cmdptr]
  long[dataptr]  := val
  long[cmdptr] := (((PIOWRITEK + (n << 9))+vaddr)<<9)|PIOWRITE
  repeat while long[cmdptr]

DAT

'----------------------------------------------------------------------------------------------------
' Start at COG address 0
'----------------------------------------------------------------------------------------------------

        org   $0

MCP23S17_COG

'----------------------------------------------------------------------------------------------------
' VMCOG initialization code - gets overwritten by vmflush, so no wasted space :)
'----------------------------------------------------------------------------------------------------

init_vm mov   pcmd,  par      ' save pointer to mailbox, first word is command
        add   pcmd,  #4
        mov   paddr, pcmd     ' paddr = pointer to long in hub that holds the addr
        add   pcmd,  #4
        mov   pdata, pcmd     ' pdata = pointer to long in hub for byte/word/long to be read/written to/from vm
        sub   pcmd,  #8

        mov   dira,SPI_DIR    ' initialize I/O direction
        movd  outa,#6

'----------------------------------------------------------------------------------------------------
' Wait for a command
'----------------------------------------------------------------------------------------------------

waitcmd wrlong zero,pcmd

wl	rdlong inst, pcmd wz  ' top 23 bits = address, bottom 9 = command
	mov   addr, inst
 if_z	jmp   #wl
	shr   addr, #9

        add   inst,#vector    ' adjust for start of vector table
vector	jmp   inst            ' ignores top 23 bits

'----------------------------------------------------------------------------------------------------
' MCP23S17_COG command vectors
'----------------------------------------------------------------------------------------------------

        jmp   #pio_read       ' vector 1
        jmp   #pio_write      ' vector 2

'----------------------------------------------------------------------------------------------------
'
' pio_read - return byte register from specified PIO
'
'----------------------------------------------------------------------------------------------------

pio_read ' movd outa,#PIO_DEVICE  ' uncomment on PropCade
        shl   addr,#16
        mov   bits,#16

        andn  outa,CS_CLK_PIN
        call  #send

        mov   data,#0
        mov   bits,#8

grp     or    outa,CLK_PIN
        test  MISO_PIN,ina wc
        rcl   data,#1
        andn  outa,CLK_PIN
        djnz  bits,#grp

        or    outa,CS_PIN
        wrlong data,pdata

        jmp   #waitcmd

'----------------------------------------------------------------------------------------------------
'
' pio_write - set byte register of specified PIO
'
'----------------------------------------------------------------------------------------------------

pio_write ' movd outa,#PIO_DEVICE  ' uncomment on PropCade
        rdlong data,pdata
        and   data,#$ff

        shl   addr,#8
        or    addr,data
        shl   addr,#8
        mov   bits,#24

        andn  outa,CS_CLK_PIN
        call  #send
        or    outa,CS_PIN

        jmp   #waitcmd

'----------------------------------------------------------------------------------------------------
' SPI send routine
'----------------------------------------------------------------------------------------------------

send    andn  outa,CLK_PIN
        rol   addr, #1 wc
        muxc  outa,MOSI_PIN
        or    outa,CLK_PIN
        djnz  bits,#send
        andn  outa,CLK_MOSI_PIN
send_ret ret

'----------------------------------------------------------------------------------------------------
' COG variables
'----------------------------------------------------------------------------------------------------

pcmd      LONG 0                ' used to hold value of PAR, points at mailbox
paddr     LONG 0                ' pointer to hub location containing addr
pdata     LONG 0                ' pointer to hub location containing byte/word/long to be read/written

addr      long 0
data      long 0
bits      long 0

inst      long 0                ' place holder for handler address
zero      long 0


CS_PIN    long 1<<CS
CLK_PIN   long 1<<CLK
MOSI_PIN  long 1<<MOSI
MISO_PIN  long 1<<MISO

CS_CLK_PIN   long (1<<CS)|(1<<CLK)
CLK_MOSI_PIN long (1<<CLK)|(1<<MOSI)

' PropCade SPI pin directions
SPI_DIR   long (1<<CS)|(1<<CLK)|(1<<MOSI)

' Mem+ SPI pin directions
'SPI_DIR   long (1<<CS)|(1<<CLK)|(1<<MOSI) | (7 << 9)

